import json
import os
import shutil
import time
import tempfile
import zipfile
from string import Template
import sys
import boto3
import importlib_resources as pkg_resources

import uuid
import requests

time_delay=2

# abstraction to return list of strings of paths of all files present in a given directory
def get_all_file_paths_in_directory(directory):
    file_paths = []
    for root, directories, files in os.walk(directory):
        for filename in files:
            filepath = os.path.join(root, filename)
            file_paths.append(filepath)
    return file_paths

# abstraction to get the details of a repository
def get_repository_details(user_session, repo_name):
    #print("Fetching details of repository \"" + repo_name + "\".")
    ecr_client = user_session.client("ecr")
    try:
        repo_details = ecr_client.describe_repositories(repositoryNames=[repo_name])['repositories']
    except:
        repo_details = []   # no details
    return repo_details

# abstraction to get the details of an image
def get_image_details(user_session, repo_name, image_tag):
    #print("Fecthing details of image \"" + repo_name + ":" + image_tag + "\".")
    ecr_client = user_session.client('ecr')
    try:
        image_details = ecr_client.describe_images(repositoryName=repo_name, imageIds=[{'imageTag': image_tag}])['imageDetails']
        #print("Fetched details of image \"" + repo_name + ":" + image_tag + "\" successfully.")
    except:
        #print("No such image \"" + repo_name + ":" + image_tag + "\" exists.")
        image_details = []
    return image_details

# abstraction to create repository with name repo_name
def create_repository(user_session, repo_name):
    #print("Creating repository \"" + repo_name + "\".")
    ecr_client = user_session.client('ecr')
    response = ecr_client.create_repository(
        repositoryName=repo_name
    )
    #print("Created repository  \"" + repo_name + "\" successfully.")

# abstraction to upload file to S3 bucket
def upload_file_to_s3(user_session, local_file_path, bucket_name, bucket_file_path):
    #print("Uploading function files to \"" + bucket_file_path +"\".")
    s3_client = user_session.client("s3")
    s3_client.upload_file(
        local_file_path,   # path to file in the local environment
        bucket_name,    # S3 bucket name
        bucket_file_path   # path to file in the S3 bucket
    )
    #print("Uploaded function files to \"" + bucket_file_path +"\" successfully.")

# abstraction to delete file from S3 bucket
def delete_file_from_s3(user_session, bucket_name, bucket_file_path):
    #print("Deleting file \"" + bucket_file_path +"\".")
    s3_client = user_session.client("s3")
    s3_client.delete_object(
        Bucket=bucket_name,    # S3 bucket name
        Key=bucket_file_path   # path to file in the S3 bucket
    )
    #print("Deleted file \"" + bucket_file_path +"\" successfully.")

# abstraction to delete IAM role
def delete_iam_role(user_session, role_name):
    #print("Deleting IAM role \"" + role_name + "\".")
    iam_client = user_session.client("iam")
    # see if role exists
    try:
        response = iam_client.get_role(RoleName=role_name)
    except:
        return
    # once role's existence is verified, fetch all attached policies
    response = iam_client.list_attached_role_policies(
        RoleName=role_name
    )
    # detach the policy from the role
    policies = response['AttachedPolicies']
    for policy in policies:
        response = iam_client.detach_role_policy(
            RoleName=role_name,
            PolicyArn=policy['PolicyArn']
        )
    # delete role
    response = iam_client.delete_role(
        RoleName=role_name
    )
    # give time to reflect in IAM
    time.sleep(time_delay)
    #print("Deleted IAM role \"" + role_name + "\" successfully.")

# abstraction to delete IAM policy
def delete_iam_policy(user_session, policy_name):

    sts_client = user_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    
    #print("Deleting IAM policy \"" + policy_name + "\".")
    iam_client = user_session.client("iam")
    policy_arn = "arn:aws:iam::" + account_id + ":policy/" + policy_name
    # see if policy exists
    try:
        response = iam_client.get_policy(PolicyArn=policy_arn)
    except:
        return
    # once policy's existence is verified, delete policy
    response = iam_client.delete_policy(
        PolicyArn=policy_arn
    )
    # give time to reflect in IAM
    time.sleep(time_delay)
    #print("Deleted IAM policy \"" + policy_name + "\" successfully.")

# abstraction to create IAM role
def create_iam_role(user_session, role_name, trust_relationship):
    #print("Creating IAM role \"" + role_name + "\".")
    iam_client = user_session.client("iam")
    response = iam_client.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(trust_relationship)     # convert JSON to string
    )
    # give time to reflect in IAM
    time.sleep(time_delay)
    #print("Created IAM role \"" + role_name + "\" successfully.")
    
# abstraction to create IAM policy
def create_iam_policy(user_session, policy_name, policy):

    sts_client = user_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    
    #print("Creating IAM policy \"" + policy_name + "\".")
    iam_client = user_session.client("iam")
    policy_arn = "arn:aws:iam::" + account_id + ":policy/" + policy_name
    response = iam_client.create_policy(
        PolicyName=policy_name,
        PolicyDocument=json.dumps(policy)     # convert JSON to string
    )
    # give time to reflect in IAM
    time.sleep(time_delay)
    #print("Created IAM policy \"" + policy_name + "\" successfully.")

# abstraction to attach IAM policy to IAM role
def attach_policy_to_role(user_session, role_name, policy_name):

    sts_client = user_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    
    #print("Attaching IAM policy \"" + policy_name +"\" to IAM role \"" + role_name + "\".")
    iam_client = user_session.client("iam")
    policy_arn = "arn:aws:iam::" + account_id + ":policy/" + policy_name
    response = iam_client.attach_role_policy(
        RoleName = role_name,
        PolicyArn = policy_arn
    )
    # give time to reflect in IAM
    time.sleep(time_delay)
    #print("Attached IAM policy \"" + policy_name +"\" to IAM role \"" + role_name + "\" successfully.")

# build image using CodeBuild from files in zip file
def build_image(user_session, bucket_name, zip_file, image_name, image):

    from . import iam

    # upload zip file to S3 bucket
    upload_file_to_s3(user_session, zip_file, bucket_name, image_name+'.zip')

    # reading JSON of the trust relationship required to create role and authorize it to use CodeBuild to build Docker image
    role_name = "codebuild_role"
    trust_relationship = json.loads(pkg_resources.read_text(iam, "codebuild_trust_relationship.txt"))
    
    # delete role for CodeBuild if role with same name exists
    delete_iam_role(user_session, role_name)

    # creating role for CodeBuild
    create_iam_role(user_session, role_name, trust_relationship)

    # reading JSON of all the policies that CodeBuild requires for accessing AWS services 
    policy_name = "codebuild_policy"
    policy = json.loads(pkg_resources.read_text(iam, "codebuild_policy.txt"))

    # delete policy for CodeBuild if policy with same name exists
    delete_iam_policy(user_session, policy_name)

    # creating policy for CodeBuild
    create_iam_policy(user_session, policy_name, policy)

    # attaching policies to role to execute CodeBuild to build Docker image
    attach_policy_to_role(user_session, role_name, policy_name)

    time.sleep(5)

    # creating CodeBuild project
    # specify which zip to be sourced from S3 that contains all the files to create the image
    # and specify the Linux environment that will be used to build the image
    codebuild_project_name = 'codebuild_' + image_name.replace("/", "_") + '_project'
    codebuild_client = user_session.client("codebuild")
    counter=1
    while(counter<=3):
        try:
            print("Attempt " + str(counter) + " to create CodeBuild project.")
            response = codebuild_client.create_project(
                name = codebuild_project_name,
                source = {
                    "type": "S3",   # where to fetch the source files from
                    "location": bucket_name + "/" + image_name + ".zip"    # exact location of the zip in S3 bucket
                },
                artifacts = {
                    "type": "S3",   # where to store the artifacts
                    "location": bucket_name   # which bucket to store artifacts in
                },
                environment = {
                    "type": "LINUX_CONTAINER",    # using a Linux environment to build the Docker image
                    "image": "aws/codebuild/standard:5.0",    # type of image to use to build Docker image
                    "computeType": "BUILD_GENERAL1_SMALL",    # compute type to use based on the user's choice
                    "privilegedMode": True    # so that a Docker image can be built inside the image
                },
                serviceRole=role_name    # role that CodeBuild will use to build project
            )
            response = codebuild_client.start_build(
                projectName = codebuild_project_name
            )
            break
        except:
            counter+=1
            if(counter<=3):
                print("CodeBuild project creation failed. Waiting for dependent resources to reflect. Retrying again in 10 seconds.")
                time.sleep(time_delay)
            else:
                print("CodeBuild project creation failed.")
                delete_file_from_s3(user_session, bucket_name, image_name+'.zip')   # delete zip file from S3 bucket
                return

    # running through loop while build status shows termination/successful completion
    counter=0
    while(True):
        build_response = codebuild_client.batch_get_builds(ids=[response['build']['id']])
        build_status = build_response['builds'][0]['buildStatus']
        if build_status == 'SUCCEEDED':
            response = codebuild_client.delete_project(     # delete CodeBuild project after process completes
                name = codebuild_project_name
            )
            print("Image successfully built.")
            print("CodeBuild finished with status " + build_status)
            print("\nNew Base Image URI: " + image)
            break
        elif build_status == 'FAILED' or build_status == 'FAULT' or build_status == 'STOPPED' or build_status == 'TIMED_OUT':
            response = codebuild_client.delete_project(     # delete CodeBuild project after process completes
                name = codebuild_project_name
            )
            print("Image not successfully built.")
            print("CodeBuild finished with status " + build_status)
            break        
        sys.stdout.write('\r')
        sys.stdout.write("Waiting" + "."*counter)
        sys.stdout.flush()
        counter=(counter+1)%4
        time.sleep(1)

    # delete zip file from S3 bucket
    delete_file_from_s3(user_session, bucket_name, image_name+'.zip')

# create a base image containing a particular set of libraries in repository with specific image tag
def build_new_base_image(libraries, repository, image_tag, python_version):

    from . import containerization_templates

    user_session = boto3.session.Session(aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
                                         aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY"),
                                         region_name=os.environ.get("AWS_REGION"))

    bucket_name = os.environ.get("BUCKET_NAME")

    s3_client = user_session.client("s3")
    
    def create_bucket(s3_client, bucket_name, region):
        try:
            response=s3_client.head_bucket(Bucket=bucket_name)
        except:
            if(region=="us-east-1"):
                response = s3_client.create_bucket(
                    ACL="private",
                    Bucket=bucket_name
                )
            else:
                location={'LocationConstraint': region}
                response=s3_client.create_bucket(
                    ACL="private",
                    Bucket=bucket_name,
                    CreateBucketConfiguration=location
                )
        return response

    create_bucket(s3_client, bucket_name, os.environ.get("AWS_REGION"))

    print("S3 Bucket \"" + bucket_name + "\" used for all storage purposes.")
    
    unique_name = repository + "_" + image_tag

    try:
        create_repository(user_session, repository)
    except:
        print("Repository already exists.")

    sts_client = user_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    region = user_session.region_name

    print("Building new base image.")

    folder_name = "unique_name"
    ##################################################
    #label = "libraries=" + ",".join(libraries)      # label of image will be all string of all libraries
    label = "libraries=test"      # label of image will be all string of all libraries
    ##################################################

    # temporary folder path where we will create all files and folder
    temp_dir = tempfile.gettempdir() + "/" + folder_name

    if(os.path.isdir(temp_dir)):
        shutil.rmtree(temp_dir)
    os.mkdir(temp_dir)

    # list of all Python libraries (with their versions if required) required to be downloaded from PyPI into Docker image
    with open(os.path.join(temp_dir, "requirements.txt"), "a") as f:
        for lib in libraries:
            f.write('%s\n' % lib)

    # Dockerfile.txt template being read and appropriate variables being assigned to generate Dockerfile
    data = pkg_resources.read_text(containerization_templates, "Dockerfile.txt")   # read template from containerization folder
    template = Template(data)
    newdata = template.substitute(
        python_version=python_version)  # AWS maintained images with speicific python versions
    with open(os.path.join(temp_dir, "Dockerfile"), "w") as file:
        file.write(newdata)

    # buildspec.txt template being read and appropriate variables being assigned to generate buildspec.yml
    data = pkg_resources.read_text(containerization_templates, "buildspec.txt")   # read template from containerization folder
    template = Template(data)
    newdata = template.substitute(
        account_id=account_id,      # AWS account id
        region=region,      # region in which the repository is / should be created
        repository=repository,      # name of the repository
        image_tag=image_tag,        # version / tag to be given to the image
        label=label)     #label of the library
    with open(os.path.join(temp_dir, "buildspec.yml"), "w") as file:
        file.write(newdata)

    # lambda_function.py being generated which has the handler that will be called when Docker image is invoked
    data = pkg_resources.read_text(containerization_templates, "lambda_function.txt")   # read template from containerization folder
    with open(os.path.join(temp_dir, "lambda_function.py"), "w") as file:
        file.write(data)

    file_paths = get_all_file_paths_in_directory(temp_dir)    # getting list of strings containing paths of all files

    # zipping all files in the temporary folder to be uploaded to the S3 bucket
    with zipfile.ZipFile(temp_dir + ".zip", "w") as zip:
        for file in file_paths:
            zip.write(file, file.replace(temp_dir, ""))      # ignore temporary file path when copying to zip file

    build_image(user_session, bucket_name, temp_dir + ".zip", unique_name + "_base_image", repository+":"+image_tag)

    if(os.path.isdir(temp_dir)):
        shutil.rmtree(temp_dir)

    return

# create lambda function using a base image from a specific repository having a specific tag
def create_lambda_using_base_image(user_session, bucket_name, directory, lambda_name, api_id, repository, image_tag, memory_size, timeout):

    from . import iam

    cloudformation_client = user_session.client('cloudformation')

    sts_client = user_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    region = user_session.region_name

    temp_dir = tempfile.gettempdir() + "/" + lambda_name

    if(os.path.isdir(temp_dir)):
        shutil.rmtree(temp_dir)

    shutil.copytree(directory, temp_dir)     # copying files from the local directory to tmp folder directory

    temp_path_directory_file_paths = get_all_file_paths_in_directory(temp_dir)    # getting list of strings containing paths of all files

    # zipping all files in a temporary folder to be uploaded to the S3 bucket
    with zipfile.ZipFile(temp_dir + ".zip", "w") as zip:    # remove temporary path from directory if present
        for file in temp_path_directory_file_paths:
            zip.write(file, file.replace(temp_dir, ""))    # ignore temporary path when copying to zip file

    upload_file_to_s3(user_session, temp_dir + ".zip", bucket_name, api_id + "/" + lambda_name + ".zip")        # upload zip file to S3 bucket

    # reading JSON of the trust relationship required to create role and authorize it to use CodeBuild to build Docker image
    role_name = "lambda_role_" + api_id
    trust_relationship = json.loads(pkg_resources.read_text(iam, "lambda_trust_relationship.txt"))
    
    # creating role for CodeBuild
    create_iam_role(user_session, role_name, trust_relationship)

    # reading JSON of all the policies that CodeBuild requires for accessing AWS services 
    policy_name = "lambda_policy_" + api_id
    policy = json.loads(pkg_resources.read_text(iam, "lambda_policy.txt"))

    # creating policy for CodeBuild    
    create_iam_policy(user_session, policy_name, policy)

    # attaching policies to role to execute CodeBuild to build Docker image
    attach_policy_to_role(user_session, role_name, policy_name)

    #print("Creating Lambda function \"" + lambda_name + "\".")

    role = 'arn:aws:iam::' + account_id + ':role/' + role_name
    code = {
        'ImageUri': account_id + '.dkr.ecr.' + region + '.amazonaws.com/' + repository + ":" + image_tag
    }
    package_type = "Image"
    timeout = int(timeout)
    memory_size = int(memory_size)
    environment = {
        'Variables': {
            'bucket': bucket_name,     # bucket where zip file is located
            'api_id': api_id,     # api_id in the bucket in which zip file is stored
            'function_name': lambda_name,        # Lambda function name
            'NUMBA_CACHE_DIR': '/tmp'
        }
    }

    lambda_client = user_session.client("lambda")
    counter=1
    while(counter<=3):
        try:
            #print("Attempt " + str(counter) + " to create Lambda function.")
            response_lambda = lambda_client.create_function(
                FunctionName=lambda_name,
                Role = role,
                Code = code,
                PackageType = package_type,
                Timeout = timeout,
                MemorySize = memory_size,
                Environment = environment
            )
            break
        except:
            counter+=1
            if(counter<=3):
                time.sleep(5)
            else:
                print("Lambda function creation failed.")

    # running through loop until function reflects
    counter=1
    while(counter<=3):
        try:
            response = lambda_client.get_function(
                FunctionName=lambda_name
            )
            #print("Created Lambda function " + lambda_name + "\" successfully.")
            break
        except:
            counter+=1
            if(counter<=3):
                #print("Lambda function did not reflected. Waiting for Lambda function to reflect.")
                time.sleep(5)
            else:
                None
                #print("Lambda function did not reflected.")

    # stack_name = 'stack-' + api_id
    # role_name = "lambda_role_" + api_id
    # trust_policy = json.loads(pkg_resources.read_text(iam, "lambda_trust_relationship.txt"))
    # policy_name = "lambda_policy_" + api_id
    # policy = json.loads(pkg_resources.read_text(iam, "lambda_policy.txt"))
    
    # role = 'arn:aws:iam::' + account_id + ':role/' + role_name
    # code = {
    #     'ImageUri': account_id + '.dkr.ecr.' + region + '.amazonaws.com/' + repository + ":" + image_tag
    # }
    # package_type = "Image"
    # timeout = int(timeout)
    # memory_size = int(memory_size)
    # environment = {
    #     'Variables': {
    #         'bucket': bucket_name,     # bucket where zip file is located
    #         'api_id': api_id,     # api_id in the bucket in which zip file is stored
    #         'function_name': lambda_name,        # Lambda function name
    #         'NUMBA_CACHE_DIR': '/tmp'
    #     }
    # }
    # arn_prefix = "arn:aws:execute-api:" + region + ":" + account_id + ":" + api_id
    # template_body = get_cloudformation_template()
    # template = Template(template_body)
    # new_template = template.substitute(
    #     PolicyDocument = json.dumps(policy),
    #     PolicyName = policy_name,
    #     TrustPolicy = json.dumps(trust_policy),
    #     Policies = json.dumps([policy_name]),
    #     RoleName = role_name,
    #     Code = json.dumps(code),
    #     Environment = json.dumps(environment),
    #     FunctionName = lambda_name,
    #     MemorySize = memory_size,
    #     PackageType = package_type,
    #     Timeout = timeout,
    #     Action1 = 'lambda:InvokeFunction',
    #     Principal1 = 'apigateway.amazonaws.com',
    #     SourceArn1 = arn_prefix + "/*/POST/m"
    # )
    # template_body = new_template

    # response = cloudformation_client.create_stack(
    #     StackName = stack_name,
    #     TemplateBody = template_body,
    #     DisableRollback=False,
    #     Capabilities=['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM', 'CAPABILITY_AUTO_EXPAND']
    # )

    os.remove(temp_dir + ".zip")    # delete the zip file created in tmp directory
    if(os.path.isdir(temp_dir)):    # delete the temporary folder created in tmp directory
        shutil.rmtree(temp_dir)

    return response

# check if the image exists in the specified repository with specified image tag
def check_if_image_exists(user_session, repo_name, image_tag):
    #print("Checking if image \"" + repo_name + ":" + image_tag +"\" exists.")
    import datetime
    ecr_client = user_session.client('ecr')
    try:
        image_details = ecr_client.describe_images(
            repositoryName=repo_name,
            imageIds=[{'imageTag': image_tag}]
        )
        result = True
        #adding date check of last image updates for any image tags.
        #if user has old image then we update it using this approach
        #result=image_details['imageDetails'][0]['imagePushedAt'].date()<=datetime.date(2021, 10, 6)

        #print("The image " + "\"" + repo_name + ":" + image_tag + "\"" + " exists.")
        # print details of image
        #print(image_details)
        return result
    except:
        #print("The image " + "\"" + repo_name + ":" + image_tag + "\"" + " does not exist.")
        return False

# check if repo exists
def check_if_repo_exists(user_session, repo_name):
    #print("Checking if repository \"" + repo_name + "\" exists.")
    ecr_client = user_session.client("ecr")
    try:
        repo_details = ecr_client.describe_images(
            repositoryName=repo_name
        )
        #print("The repository \"" + repo_name + "\" exists.")
        # print details all images in repo
        #print(repo_details)
        return True
    except:
        #print("The repository \"" + repo_name + "\" does not exist.")
        return False

# delete statement with specific Sid in ECR
def delete_registry_permission(user_session, statement_id):
    ecr_client = user_session.client("ecr")
    try:
        registry_policy = ecr_client.get_registry_policy()
        policy_json = json.loads(registry_policy['policyText'])
        for statement in policy_json["Statement"]:
            if(statement["Sid"]==statement_id):
                policy_json["Statement"].remove(statement)
                response = ecr_client.put_registry_policy(json.dumps(policy_json))
                break
        time.sleep(5)
    except:
        None

# add statement with specific Sid in ECR
def add_registry_permission(user_session, statement_id, source_account_id, account_id, region):
    ecr_client = user_session.client("ecr")
    policy=json.dumps({
        "Version":"2012-10-17",
        "Statement":[
            {
                "Sid": statement_id,
                "Principal": {
                    "AWS": "arn:aws:iam::" + source_account_id + ":root"
                },
                "Action": "ecr:*",
                "Resource": [ "arn:aws:ecr:"+region+":"+account_id+":repository/*"],
                "Effect": "Allow"
            }
        ]
    })
    response = ecr_client.put_registry_policy(policyText=policy)
    time.sleep(5)

# create base image having repository:image_tag identity from source account if image not present, update base image if update flag set to true
def clone_base_image(user_session, repository, image_tag, source_account_id, api_endpoint="", update=False):

    if(check_if_image_exists(user_session, repository, image_tag) and update==False):
        result = {"Status": 1, "Success" : "Base image \"" + repository + ":" + image_tag + "\" already exists on this user's account."}
    elif(len(api_endpoint)>0):
        sts_client = user_session.client("sts")

        account_id = sts_client.get_caller_identity()["Account"]
        region = user_session.region_name

        ecr_client = user_session.client('ecr')

        statement_id = source_account_id + "_image_clone_access"

        delete_registry_permission(user_session, statement_id)
        add_registry_permission(user_session, statement_id, source_account_id, account_id, region)

        bodydata = {
            "useraccountnumber": str(account_id),  # change this to first and last name
            "region": region,
            "repository": repository,
            "image_tag": image_tag
        }
        
        # datasets api
        headers_with_authentication = {'Content-Type': 'application/json',
                                       'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"),
                                       'Access-Control-Allow-Headers': 'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization',
                                       'Access-Control-Allow-Origin': '*'}

        # modeltoapi lambda function invoked through below url to return new prediction api in response
        response = requests.post(api_endpoint, json=bodydata, headers=headers_with_authentication)
        time.sleep(5)

        # Delete registry policy
        delete_registry_permission(user_session, statement_id)

        #client.delete_registry_policy()
        result = {"Status": 1, "Success" : "New base image loaded to user's account"}

        #print(result)
    else:
        result = {"Status": 0, "Success" : "API endpoint not valid/not provided and image does not exist either."}
    return result


def get_cloudformation_template():
    return """
        {
            "AWSTemplateFormatVersion": "2010-09-09",
            "Description": "Creation of Role, Policy, Lambda, Lambda Permission",
            "Resources": {
                "Role": {
                    "Type" : "AWS::IAM::Role",
                    "Properties" : {
                        "AssumeRolePolicyDocument" : $TrustPolicy,
                        "RoleName" : "$RoleName"
                    }
                },
                "Policy": {
                    "DependsOn": "Role",
                    "Type" : "AWS::IAM::Policy",
                    "Properties" : {
                        "PolicyDocument" : $PolicyDocument,
                        "PolicyName" : "$PolicyName",
                        "Roles": ["$RoleName"]
                    }
                },
                "Lambda": {
                    "DependsOn": "Policy",
                    "Type" : "AWS::Lambda::Function",
                    "Properties" : {
                        "Code" : $Code,
                        "Environment" : $Environment,
                        "FunctionName" : "$FunctionName",
                        "MemorySize" : $MemorySize,
                        "PackageType" : "$PackageType",
                        "Role" : {
                            "Fn::GetAtt": [
                                "Role",
                                "Arn"
                            ]
                        },
                        "Timeout" : $Timeout
                    }
                },
                "LambdaInvokePermission1": {
                    "DependsOn": "Lambda",
                    "Type" : "AWS::Lambda::Permission",
                    "Properties" : {
                        "Action" : "$Action1",
                        "FunctionName" : "$FunctionName",
                        "Principal" : "$Principal1",
                        "SourceArn" : "$SourceArn1"
                    }
                }
            }
        }
    """


# import packages 




class ModelPlayground:
    """
    Parameters:
    ----------
    `model_type` : ``string``
          values - [ 'text' , 'image' , 'tabular' , 'video', 'audio','timeseries' ] 
          type of model data     
    `classification`:    ``bool, default=True``
        True [DEFAULT] if model is of Classification type with categorical target variables
        False if model is of Regression type with continuous target variables
    `private` :   ``bool, default = False``
        True if model and its corresponding data is not public
        False [DEFAULT] if model and its corresponding data is public 
    """
    def __init__(self, model_type=None, classification=None, private=None, playground_url=None):
        # confirm correct args are provided
        if playground_url != None or all([model_type !=None, classification !=None, private!=None]):
            pass
        elif playground_url == None and any([model_type ==None, classification ==None, private==None]):
            return print("Error. To instantiate a ModelPlayground instance, please provide either a playground_url or \n the model_type, classification, and private arguments.")
        
        self.model_type = model_type
        self.categorical = classification 
        self.private = private
        self.playground_url = playground_url
    
    
    def __str__(self):
        return f"ModelPlayground instance of model type: {self.model_type}, classification: {self.categorical},  private: {self.private}"
    

    def deploy(self, model_filepath, preprocessor_filepath, y_train, example_data=None, custom_libraries = "FALSE", image="", reproducibility_env_filepath=None, memory=None, timeout=None, email_list=[], pyspark_support=False):

        """
        Launches a live prediction REST API for deploying ML models using model parameters and user credentials, provided by the user
        Inputs : 7
        Output : model launched to an API
                detailed API info printed out

        Parameters: 
        ----------
        `model_filepath` :  ``string`` ends with '.onnx'
              value - Absolute path to model file 
              [REQUIRED] to be set by the user
              .onnx is the only accepted model file extension
              "example_model.onnx" filename for file in directory.
              "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
        `preprocessor_filepath`:  ``string``
            value - absolute path to preprocessor file 
            [REQUIRED] to be set by the user
            "./preprocessor.zip" 
            searches for an exported zip preprocessor file in the current directory
            file is generated using export_preprocessor function from the AI Modelshare library  
        `y_train` : training labels for classification models.
              [REQUIRED] for classification type models
              expects pandas dataframe of one hot encoded y train data
        `custom_libraries`:   ``string``
            "TRUE" if user wants to load custom Python libraries to their prediction runtime
            "FALSE" if user wishes to use AI Model Share base libraries including latest versions of most common ML libs.
        `example_data`:  pandas DataFrame (for tabular & text data) OR filepath as string (image, audio, video data)
              tabular data - pandas DataFrame in same structure expected by preprocessor function
              other data types - absolute path to folder containing example data
                                (first five files with relevent file extensions will be accepted)
                     
              [REQUIRED] for tabular data
          `email_list`: ``list of string values``
                values - list including all emails of users who have access the playground.
                list should contain same emails that were used to sign up for modelshare.org account.
                [OPTIONAL] set by the playground owner
        Returns:
        --------
        print_api_info : prints statements with generated live prediction API details
                        also prints steps to update the model submissions by the user/team
        """
        from aimodelshare.generatemodelapi import model_to_api
        self.playground_url = model_to_api(model_filepath=model_filepath, 
                                      model_type = self.model_type, 
                                      private = self.private, 
                                      categorical = self.categorical,
                                      y_train = y_train, 
                                      preprocessor_filepath = preprocessor_filepath, 
                                      example_data = example_data,
                                      custom_libraries = custom_libraries,
                                      image=image,
                                      reproducibility_env_filepath = reproducibility_env_filepath,
                                      memory=memory,
                                      timeout=timeout,
                                      email_list=email_list,
                                      pyspark_support=pyspark_support)
        #remove extra quotes
        self.playground_url = self.playground_url[1:-1]
    
    def create_competition(self, data_directory, y_test, eval_metric_filepath=None, email_list = [], public=False, public_private_split=0.5):
        """
        Creates a model competition for a deployed prediction REST API
        Inputs : 4
        Output : Create ML model competition and allow authorized users to submit models to resulting leaderboard/competition
        
        Parameters:
        -----------
        `y_test` :  ``list`` of y values for test data used to generate metrics from predicted values from X test data submitted via the submit_model() function
            [REQUIRED] to generate eval metrics in competition leaderboard
                                
        `data_directory` : folder storing training data and test data (excluding Y test data)
        `email_list`: [OPTIONAL] list of comma separated emails for users who are allowed to submit models to competition.  Emails should be strings in a list.
        `public`: [REQUIRED] True/false. Defaults to False.  If True, competition is public and ANY AIMODELSHARE USER CAN SUBMIT MODELS.  USE WITH CAUTION b/c one model and 
            one preprocessor file will be be saved to your AWS S3 folder for each model submission.
        
        
        Returns:
        -----------
        finalmessage : Information such as how to submit models to competition
        
        """
        from aimodelshare.generatemodelapi import create_competition

        competition = create_competition(self.playground_url, 
                                    data_directory, 
                                    y_test, 
                                    eval_metric_filepath,
                                    email_list, 
                                    public,
                                    public_private_split)
        return competition
        
    def create_experiment(self, data_directory, y_test, eval_metric_filepath=None, email_list = [], public=False, public_private_split=0.5):
        """
        Creates an experiment for a deployed prediction REST API
        Inputs : 4
        Output : Create ML model experiment and allows authorized users to submit models to resulting experiment tracking leaderboard
        
        Parameters:
        -----------
        `y_test` :  ``list`` of y values for test data used to generate metrics from predicted values from X test data submitted via the submit_model() function
            [REQUIRED] to generate eval metrics in experiment leaderboard
                                
        `data_directory` : folder storing training data and test data (excluding Y test data)
        `email_list`: [OPTIONAL] list of comma separated emails for users who are allowed to submit models to experiment leaderboard.  Emails should be strings in a list.
        `public`: [REQUIRED] True/false. Defaults to False.  If True, experiment is public and ANY AIMODELSHARE USER CAN SUBMIT MODELS.  USE WITH CAUTION b/c one model and 
            one preprocessor file will be be saved to your AWS S3 folder for each model submission.
        
        
        Returns:
        -----------
        finalmessage : Information such as how to submit models to experiment
        
        """
        from aimodelshare.generatemodelapi import create_experiment

        experiment = create_experiment(self.playground_url, 
                                    data_directory, 
                                    y_test, 
                                    eval_metric_filepath,
                                    email_list, 
                                    public,
                                    public_private_split)
        return experiment

    def submit_model(self, model_filepath, preprocessor_filepath, prediction_submission):
        """
        Submits model/preprocessor to machine learning experiment leaderboard and model architecture database using live prediction API url generated by AI Modelshare library
        The submitted model gets evaluated and compared with all existing models and a leaderboard can be generated 
        
        Parameters:
        -----------
        `model_filepath`:  ``string`` ends with '.onnx'
            value - Absolute path to model file [REQUIRED] to be set by the user
            .onnx is the only accepted model file extension
            "example_model.onnx" filename for file in directory.
            "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
        `preprocessor_filepath`:   ``string``, default=None
            value - absolute path to preprocessor file 
            [REQUIRED] to be set by the user
            "./preprocessor.zip" 
            searches for an exported zip preprocessor file in the current directory
            file is generated from preprocessor module using export_preprocessor function from the AI Modelshare library 
        `prediction_submission`: [REQUIRED] list of predictions from X test data that will be used to evaluate model prediction error against y test data.
            Use mycompetition.inspect_y_test() to view example of list expected by competition.
        
        Returns:
        -------
        response:   Model version if the model is submitted sucessfully
                    error  if there is any error while submitting models
        
        """

        from aimodelshare.model import submit_model as submit
        submission = submit(model = model_filepath, 
                            apiurl = self.playground_url,
                            prediction_submission = prediction_submission, 
                            preprocessor = preprocessor_filepath)
        return submission
        

    
    def update_runtime_model(self, model_version=None, submission_type="competition"):
        """
        Updates the prediction API behind the Model Playground with a new model from the leaderboard and verifies Model Playground performance metrics.

        Parameters:
        -----------
        `model_version`: ``int``
            model version number from competition leaderboard

        Returns:
        --------
        response:   success message when the model and preprocessor are updated successfully
        
        """
        from aimodelshare.model import update_runtime_model as update
        update = update(apiurl = self.playground_url, model_version = model_version, submission_type=submission_type)
        return update
        
    def instantiate_model(self, version=None, trained=False, reproduce=False, submission_type="competition"): 
        """
        Import a model previously submitted to a leaderboard to use in your session

        Parameters:
        -----------
        `version`: ``int``
            Model version number from competition or experiment leaderboard
        `trained`: ``bool, default=False``
            if True, a trained model is instantiated, if False, the untrained model is instantiated

        Returns:
        --------
        model: model chosen from leaderboard
        """
        from aimodelshare.aimsonnx import instantiate_model
        model = instantiate_model(apiurl=self.playground_url, trained=trained, version=version, reproduce=reproduce, submission_type=submission_type)
        return model
    
    def delete_deployment(self, playground_url=None):
        """
        Delete all components of a Model Playground, including: AWS s3 bucket & contents,
        attached competitions, prediction REST API, and interactive Model Playground web dashboard.

        Parameters:
        -----------
        `playground_url`: ``string`` of API URL the user wishes to delete
        WARNING: User must supply high-level credentials in order to delete an API.

        Returns:
        --------
        Success message when deployment is deleted.
        """
        from aimodelshare.api import delete_deployment
        if playground_url == None:
            playground_url = self.playground_url
        deletion = delete_deployment(apiurl = playground_url)
        return deletion

    def import_reproducibility_env(self):
        from aimodelshare.reproducibility import import_reproducibility_env_from_model
        import_reproducibility_env_from_model(apiurl=self.playground_url)

    def update_access_list(self, email_list=[], update_type="Replace_list"):
        """
        Updates list of authenticated participants who can submit new models to a competition.

        Parameters:
        -----------
        `apiurl`: string
                URL of deployed prediction API 
          
        `email_list`: [REQUIRED] list of comma separated emails for users who are allowed to submit models to competition.  Emails should be strings in a list.
        `update_type`:[REQUIRED] options, ``string``: 'Add', 'Remove', 'Replace_list','Get. Add appends user emails to original list, Remove deletes users from list, 
                  'Replace_list' overwrites the original list with the new list provided, and Get returns the current list.    

        Returns:
        --------
        response:   "Success" upon successful request
        """
        from aimodelshare.generatemodelapi import update_access_list as update_list
        update = update_list(apiurl = self.playground_url, email_list=email_list, update_type=update_type)
        return update


class Competition:
    """
    Parameters:
    ----------
    `playground_url`: playground_url attribute of ModelPlayground class or ``string``
        of existing ModelPlayground URL
    """

    submission_type = "competition"

    def __init__(self, playground_url):
        self.playground_url = playground_url
    
    def __str__(self):
        return f"Competition class instance for playground: {self.playground_url}"
        
    def submit_model(self, model_filepath, preprocessor_filepath, prediction_submission, 
        sample_data=None, reproducibility_env_filepath=None, custom_metadata=None):
        """
        Submits model/preprocessor to machine learning competition using live prediction API url generated by AI Modelshare library
        The submitted model gets evaluated and compared with all existing models and a leaderboard can be generated
        
        Parameters:
        -----------
        `model_filepath`:  ``string`` ends with '.onnx'
            value - Absolute path to model file [REQUIRED] to be set by the user
            .onnx is the only accepted model file extension
            "example_model.onnx" filename for file in directory.
            "/User/xyz/model/example_model.onnx" absolute path to model file from local directory
        `prediction_submission`:   one hot encoded y_pred
            value - predictions for test data
            [REQUIRED] for evaluation metrics of the submitted model
        `preprocessor_filepath`:   ``string``, default=None
            value - absolute path to preprocessor file 
            [REQUIRED] to be set by the user
            "./preprocessor.zip" 
            searches for an exported zip preprocessor file in the current directory
            file is generated from preprocessor module using export_preprocessor function from the AI Modelshare library 
 
        Returns:
        --------
        response:   Model version if the model is submitted sucessfully
        """

        from aimodelshare.model import submit_model as submit
        submission = submit(model_filepath = model_filepath, 
                              apiurl = self.playground_url,
                              prediction_submission = prediction_submission, 
                              preprocessor = preprocessor_filepath,
                              reproducibility_env_filepath = reproducibility_env_filepath,
                              custom_metadata = custom_metadata, 
                              submission_type = self.submission_type)
        return submission

    def instantiate_model(self, version=None, trained=False, reproduce=False): 
        """
        Import a model previously submitted to the competition leaderboard to use in your session

        Parameters:
        -----------
        `version`: ``int``
            Model version number from competition leaderboard
        `trained`: ``bool, default=False``
            if True, a trained model is instantiated, if False, the untrained model is instantiated
       
        Returns:
        --------
        model: model chosen from leaderboard
        """
        from aimodelshare.aimsonnx import instantiate_model
        model = instantiate_model(apiurl=self.playground_url, trained=trained, version=version, 
            reproduce=reproduce, submission_type=self.submission_type)
        return model

    def inspect_model(self, version=None, naming_convention=None):
        """
        Examine structure of model submitted to a competition leaderboard

        Parameters:
        ----------
        `version` : ``int``
            Model version number from competition leaderboard
      
        Returns:
        --------
        inspect_pd : dictionary of model summary & metadata
        """
        from aimodelshare.aimsonnx import inspect_model

        inspect_pd = inspect_model(apiurl=self.playground_url, version=version, 
            naming_convention=naming_convention, submission_type = self.submission_type)

        return inspect_pd

    def compare_models(self, version_list="None", by_model_type=None, best_model=None, verbose=1, naming_convention=None):
        """
        Compare the structure of two or more models submitted to a competition leaderboard.
        Use in conjuction with stylize_compare to visualize data. 
        
        Parameters:
        -----------
        `version_list` = ``list of int``
            list of model version numbers to compare (previously submitted to competition leaderboard) 
        `verbose` = ``int``
            controls the verbosity: the higher, the more detail 
        
        Returns:
        --------
        data : dictionary of model comparison information
        """
        from aimodelshare.aimsonnx import compare_models as compare
        data = compare(apiurl = self.playground_url, 
                      version_list = version_list, 
                      by_model_type = by_model_type,
                      best_model = best_model, 
                      verbose = verbose,
                      naming_convention=naming_convention,
                      submission_type = self.submission_type)
        return data

    def stylize_compare(self, compare_dict, naming_convention=None):
        """
        Stylizes data received from compare_models to highlight similarities & differences.

        Parameters:
        -----------
        `compare_dict` = dictionary of model data from compare_models
 
        Returns:
        --------
        formatted table of model comparisons 
        """
        from aimodelshare.aimsonnx import stylize_model_comparison
        stylized_compare = stylize_model_comparison(comp_dict_out=compare_dict, naming_convention=naming_convention)
        return(stylized_compare)

    def inspect_y_test(self):
        """
        Examines structure of y-test data to hep users understand how to submit models to the competition leaderboad.

        Parameters:
        ------------
        None
 
        Returns:
        --------
        dictionary of a competition's y-test metadata
        """
        from aimodelshare.aimsonnx import inspect_y_test
        data = inspect_y_test(apiurl = self.playground_url, submission_type=self.submission_type)
        return data
    
    def get_leaderboard(self, verbose=3, columns=None):
        """
        Get current competition leaderboard to rank all submitted models.
        Use in conjuction with stylize_leaderboard to visualize data. 
        
        Parameters:
        -----------
        `verbose` : optional, ``int``
            controls the verbosity: the higher, the more detail 
        `columns` : optional, ``list of strings``
            list of specific column names to include in the leaderboard, all else will be excluded
            performance metrics will always be displayed
        
        Returns:
        --------
        dictionary of leaderboard data 
        """
        from aimodelshare.leaderboard import get_leaderboard
        data = get_leaderboard(verbose=verbose,
                 columns=columns, 
                 apiurl = self.playground_url, 
                 submission_type=self.submission_type)
        return data
    
    def stylize_leaderboard(self, leaderboard, naming_convention="keras"):
        """
        Stylizes data received from get_leaderbord.

        Parameters:
        -----------
        `leaderboard` : data dictionary object returned from get_leaderboard

        Returns:
        --------
        Formatted competition leaderboard
        """
        from aimodelshare.leaderboard import stylize_leaderboard as stylize_lead
        stylized_leaderboard = stylize_lead(leaderboard = leaderboard, naming_convention=naming_convention)
        return stylized_leaderboard
    
    def update_access_list(self, email_list=[],update_type="Replace_list"):
        """
        Updates list of authenticated participants who can submit new models to a competition.

        Parameters:
        -----------
        `apiurl`: string
                URL of deployed prediction API 
          
        `email_list`: [REQUIRED] list of comma separated emails for users who are allowed to submit models to competition.  Emails should be strings in a list.
        `update_type`:[REQUIRED] options, ``string``: 'Add', 'Remove', 'Replace_list','Get. Add appends user emails to original list, Remove deletes users from list, 
                  'Replace_list' overwrites the original list with the new list provided, and Get returns the current list.    

        Returns:
        --------
        response:   "Success" upon successful request
        """
        from aimodelshare.generatemodelapi import update_access_list as update_list
        update = update_list(apiurl = self.playground_url, 
            email_list=email_list,update_type=update_type,
            submission_type=self.submission_type)
        return update



class Experiment(Competition):
    """
    Parameters:
    ----------
    `playground_url`: playground_url attribute of ModelPlayground class or ``string``
        of existing ModelPlayground URL
    """

    submission_type = "experiment"

    def __init__(self, playground_url):
        self.playground_url = playground_url
    
    def __str__(self):
        return f"Experiment class instance for playground: {self.playground_url}"
        
    

class Data: 
    def __init__(self, data_type, playground_url=None):
        self.data_type = data_type
        self.playground_url = playground_url 
    
    def __str__(self):
        return f"This is a description of the Data class."

    def share_dataset(self, data_directory="folder_file_path", classification="default", private="FALSE"): 
        from aimodelshare.data_sharing.share_data import share_dataset as share
        response = share(data_directory=data_directory, classification=classification, private=private)
        return response
    
    def download_data(self, repository):
        from aimodelshare.data_sharing.download_data import download_data as download
        datadownload = download(repository)
        return datadownload




import os
import shutil
import tempfile

def delete_files_from_temp_dir(temp_dir_file_deletion_list):
    temp_dir = tempfile.gettempdir()
    for file_name in temp_dir_file_deletion_list:
        file_path = os.path.join(temp_dir, file_name)
        if(os.path.exists(file_path)):
            os.remove(file_path)

def delete_folder(folder_path):
    if os.path.exists(folder_path):
        shutil.rmtree(folder_path)

def make_folder(folder_path):
    os.makedirs(folder_path, exist_ok=True)


import os
import sys
import json
import random
import tempfile
import pkg_resources

import numpy as np
import tensorflow as tf

from aimodelshare.aws import get_s3_iam_client, run_function_on_lambda, get_aws_client

def export_reproducibility_env(seed, directory, mode="gpu"):
  # Change the output into json.dumps
  # Argument single seed for all inputs & mode
  data = {
    "global_seed_code": [
      "os.environ['PYTHONHASHSEED'] = '{}'".format(seed),
      "random.seed({})".format(seed),
      "tf.random.set_seed({})".format(seed),
      "np.random.seed({})".format(seed),
    ]
  }

  # Ignore this part for now
  # Local seed codes are tensorflow code that are
  # not affected by the global seed and sometimes require us
  # to define the seed in the function call
  data["local_seed_code"] = [
    "train_ds = tf.keras.preprocessing.image_dataset_from_directory(data_dir, validation_split=0.2, subset='training', seed={}, image_size=(img_height, img_width), batch_size=batch_size)".format(seed),
    "val_ds = tf.keras.preprocessing.image_dataset_from_directory(data_dir, validation_split=0.2, subset='validation', seed={}, image_size=(img_height, img_width), batch_size=batch_size)".format(seed),
  ]

  if mode == "gpu":
    data["gpu_cpu_parallelism_ops"] = [
      "os.environ['TF_DETERMINISTIC_OPS'] = '1'",
      "os.environ['TF_CUDNN_DETERMINISTIC'] = '1'"
    ]
  elif mode == "cpu":
    data["gpu_cpu_parallelism_ops"] = [
      "tf.config.threading.set_inter_op_parallelism_threads(1)"
    ]
  else:
    raise Exception("Error: unknown 'mode' value, expected 'gpu' or 'cpu'")

  installed_packages = pkg_resources.working_set
  installed_packages_list = sorted(["%s==%s" % (i.key, i.version)
    for i in installed_packages])

  data["session_runtime_info"] = {
    "installed_packages": installed_packages_list,
    "python_version": sys.version,
  }

  with open(os.path.join(directory, "reproducibility.json"), "w") as fp:
    json.dump(data, fp)

  return print("Your reproducibility environment is now saved to 'reproducibility.json'")

def set_reproducibility_env(reproducibility_env):
  # Change the input into dict / json
  for global_code in reproducibility_env["global_seed_code"]:
    exec("%s" % (global_code))

  for parallelism_ops in reproducibility_env["gpu_cpu_parallelism_ops"]:
    exec("%s" % (parallelism_ops))

def import_reproducibility_env(reproducibility_env_file):
  with open(reproducibility_env_file) as json_file:
    reproducibility_env = json.load(json_file)
    set_reproducibility_env(reproducibility_env)

  print("Your reproducibility environment is successfully setup")

def import_reproducibility_env_from_model(apiurl):
    if all(["AWS_ACCESS_KEY_ID" in os.environ, 
            "AWS_SECRET_ACCESS_KEY" in os.environ,
            "AWS_REGION" in os.environ, 
            "username" in os.environ, 
            "password" in os.environ]):
        pass
    else:
        return print("'Instantiate Model' unsuccessful. Please provide credentials with set_credentials().")

    aws_client = get_aws_client()
    reproducibility_env_filename = "/runtime_reproducibility.json"

    # Get bucket and model_id for user
    response, error = run_function_on_lambda(
        apiurl, **{"delete": "FALSE", "versionupdateget": "TRUE"}
    )
    if error is not None:
        raise error

    _, bucket, model_id = json.loads(response.content.decode("utf-8"))

    try:
        resp_string = aws_client["client"].get_object(
            Bucket=bucket, Key=model_id + reproducibility_env_filename
        )

        reproducibility_env_string = resp_string['Body'].read()

    except Exception as err:
        print("This model was not deployed with reproducibility support")
        raise err

    # generate tempfile for onnx object 
    temp_dir = tempfile.gettempdir()
    temp_path = os.path.join(temp_dir, 'temp_file_name')

    # save onnx to temporary path
    with open(temp_path, "wb") as f:
        f.write(reproducibility_env_string)

    import_reproducibility_env(temp_path)



import boto3
import os
import requests
import uuid
import json
import math
import time
import datetime
import onnx
import tempfile
import shutil
import sys
import pickle
import tempfile

from aimodelshare.tools import extract_varnames_fromtrainingdata, _get_extension_from_filepath
from aimodelshare.aws import get_s3_iam_client, run_function_on_lambda
from aimodelshare.bucketpolicy import _custom_upload_policy
from aimodelshare.exceptions import AuthorizationError, AWSAccessError, AWSUploadError
from aimodelshare.api import create_prediction_api
from aimodelshare.api import get_api_json

from aimodelshare.preprocessormodules import upload_preprocessor
from aimodelshare.model import _get_predictionmodel_key, _extract_model_metadata

# what is private here
# what if deployment_dir is not present? Not handled yet
# Return section and private parameter documentation left
# output_list_exampledata list being accessed by which key in response?
# create file_objects in temp_dir, abstract from user
# convert example output from list to json, in frontend display value of key of result

def deploy_custom_lambda(input_json_exampledata, output_json_exampledata, lambda_filepath, deployment_dir, private, custom_libraries):

    """
        Deploys an AWS Lambda function based on the predict() function specified in the lambda_filepath .py file
        Inputs : 6
        Output : Information about the deployed API

        -----------

        Parameters :

        input_json_exampledata : JSON object [REQUIRED]
                                 JSON object representing the structure of input that Lambda expects to receive

        output_json_exampledata : List of element(s) [REQUIRED]
                                  List of element(s) representing the output that the Lambda will return
        
        lambda_filepath : String [REQUIRED]
                          Expects relative/absolute path to the .py file containing the predict() function,
                          imports to all other custom libraries that are defined by the user and used by the
                          predict() function can be placed in the deployment_dir directory
        
        deployment_dir : String
                         Expects relative/absolute path to the directory containing all the files being used by the
                         predict() function in the lambda_filepath .py file

        private : string

        custom_libraries : String of libraries ("library_1,library_2")
                           Expects a list of strings denoting libraries required for Lambda to work
                           Strings must be libraries present in PyPi
                           Installation will follow pattern - pip install <library_name>

        -----------

        Returns

        api_info : prints statements with generated live prediction API details
                   also prints steps to update the model submissions by the user/team
    """

    temp_dir = tempfile.gettempdir()

    file_objects_folder_path = os.path.join(temp_dir, 'file_objects')

    # if 'file_objects' is not the name of deployment directory deployment_dir, 'file_objects' directory
    # is created and contents of the deployment_dir directory are copied to 'file_objects' directory
    if deployment_dir != file_objects_folder_path:
        if os.path.exists(file_objects_folder_path):
            shutil.rmtree(file_objects_folder_path)
        shutil.copytree(deployment_dir, file_objects_folder_path)

    # if 'custom_lambda.py' is not the name of the custom lambda .py file lambda_filepath, 'custom_lambda.py' file
    # is created and contents of lambda_filepath .py file is written into 'custom_lambda.py'
    if lambda_filepath != 'custom_lambda.py':
        with open(lambda_filepath, 'r') as in_f:
            with open('custom_lambda.py', 'w') as out_f:
                out_f.write(in_f.read())

    # create json and upload to API folder in S3 for displaying
    json_exampledata = {
        "input_json_exampledata": input_json_exampledata,
        "output_json_exampledata": output_json_exampledata
    }

    with open(os.path.join(temp_dir, 'exampledata.json'), 'w') as f:
        json.dump(json_exampledata, f)

    aws_access_key_id = str(os.environ.get("AWS_ACCESS_KEY_ID"))
    aws_secret_access_key = str(os.environ.get("AWS_SECRET_ACCESS_KEY"))
    region_name = str(os.environ.get("AWS_REGION"))

    ### COMMENTS - TO DO
    api_json= get_api_json()        # why is this required
    user_client = boto3.client(     # creating apigateway client
        'apigateway',
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name
    )
    response2 = user_client.import_rest_api(        # what is being imported
        failOnWarnings = True,
        parameters = {'endpointConfigurationTypes': 'REGIONAL'},
        body = api_json
    )
    ###

    start = datetime.datetime.now()

    api_id = response2['id']
    now = datetime.datetime.now()
    s3, iam, region = get_s3_iam_client(aws_access_key_id, aws_secret_access_key, region_name)

    s3["client"].create_bucket(
        Bucket=os.environ.get("BUCKET_NAME")
    )

    apiurl = create_prediction_api(None, str(api_id), 'custom', 'FALSE', [], api_id, "TRUE", custom_libraries)

    print("\n\nWe need some information about your model before we can generate your API.\n")
    aishare_modelname = input("Name your deployment: ")
    aishare_modeldescription = input("Describe your deployment: ")
    aishare_modelevaluation = input("Describe your deployment's performance (OPTIONAL): ")
    aishare_tags = input("Enter comma-separated search categories for your deployment (OPTIONAL): ")
    aishare_apicalls = 0
    print('')
    # unpack user credentials
    unique_model_id = str(api_id)
    bucket_name = os.environ.get("BUCKET_NAME")

    # why is this being done here, can it not be abstracted

    categorical="FALSE" # categorical is being used where, for what
    bodydata = {
        "id": int(math.log(1/((time.time()*1000000)))*100000000000000),
        "unique_model_id": unique_model_id,
        "apideveloper": os.environ.get("username"),  # change this to first and last name
        "apimodeldescription": aishare_modeldescription,
        "apimodelevaluation": aishare_modelevaluation,
        "apimodeltype": 'custom',
        # getting rid of extra quotes that screw up dynamodb string search on apiurls
        "apiurl": apiurl['body'].strip('\"'),
        "bucket_name": bucket_name,
        "version": 1,
        "modelname": aishare_modelname,
        "tags": aishare_tags,
        "Private": private,
        "Categorical": categorical,
        "delete": "FALSE",
    }

    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                  json=bodydata, headers=headers_with_authentication)

    # Get the response
    headers_with_authentication = {'Content-Type': 'application/json', 'authorizationToken': os.environ.get("JWT_AUTHORIZATION_TOKEN"), 'Access-Control-Allow-Headers':
                                   'Content-Type,X-Amz-Date,authorizationToken,Access-Control-Allow-Origin,X-Api-Key,X-Amz-Security-Token,Authorization', 'Access-Control-Allow-Origin': '*'}
    # modeltoapi lambda function invoked through below url to return new prediction api in response
    response = requests.post("https://bhrdesksak.execute-api.us-east-1.amazonaws.com/dev/modeldata",
                              json=bodydata, headers=headers_with_authentication)
    response_string = response.text
    response_string = response_string[1:-1]

    end = datetime.datetime.now()    # end timer
    difference = (end - start).total_seconds()
    finalresult2 = "Your AI Model Share API was created in " + \
        str(int(difference)) + " seconds." + " API Url: " + apiurl['body']
    s3, iam, region = get_s3_iam_client(os.environ.get("AWS_ACCESS_KEY_ID"), os.environ.get("AWS_SECRET_ACCESS_KEY"), os.environ.get("AWS_REGION"))
    policy_response = iam["client"].get_policy(
        PolicyArn=os.environ.get("POLICY_ARN")
    )
    user_policy = iam["resource"].UserPolicy(
        os.environ.get("IAM_USERNAME"), policy_response['Policy']['PolicyName'])
    response = iam["client"].detach_user_policy(
        UserName= os.environ.get("IAM_USERNAME"),
        PolicyArn=os.environ.get("POLICY_ARN")
    )
    # add new policy that only allows file upload to bucket
    policy = iam["resource"].Policy(os.environ.get("POLICY_ARN"))
    response = policy.delete()
    s3upload_policy = _custom_upload_policy(bucket_name, unique_model_id)
    s3uploadpolicy_name = 'temporaryaccessAImodelsharePolicy' + \
        str(uuid.uuid1().hex)
    s3uploadpolicy_response = iam["client"].create_policy(
        PolicyName=s3uploadpolicy_name,
        PolicyDocument=json.dumps(s3upload_policy)
    )
    user = iam["resource"].User(os.environ.get("IAM_USERNAME"))
    response = user.attach_policy(
        PolicyArn=s3uploadpolicy_response['Policy']['Arn']
    )
    finalresultteams3info = "Your team members can submit improved models to your prediction api using the update_model_version() function. \nTo upload new models and/or preprocessors to this model team members should use the following awskey/password/region:\n\n aws_key = " + \
        os.environ.get("AI_MODELSHARE_ACCESS_KEY_ID") + ", aws_password = " + os.environ.get("AI_MODELSHARE_SECRET_ACCESS_KEY") + " region = " + \
        os.environ.get("AWS_REGION") +".  \n\nThis aws key/password combination limits team members to file upload access only."
    api_info = finalresult2+"\n"

    # Build output {{{
    final_message = ("Follow this link to explore your Model Playground's functionality\n"
                     "You can make predictions with the cURL functionality and access example code from the Programmatic tab.\n")
    web_dashboard_url = ("https://www.modelshare.org/detail/"+ response_string)

    end = datetime.datetime.now()
    difference = (end - start).total_seconds()
    finalresult2 = "Success! Your Model Playground was created in " + \
        str(int(difference)) + " seconds."

    print(api_info)

    print("\n\n" + final_message + web_dashboard_url)
    
    return

__all__ = [
    deploy_custom_lambda
]



from .object_oriented import ModelPlayground, Competition, Data
from .preprocessormodules import export_preprocessor,upload_preprocessor,import_preprocessor
from .data_sharing.download_data import download_data, import_quickstart_data
from .reproducibility import export_reproducibility_env, import_reproducibility_env


__all__ = [
    # Object Oriented
    ModelPlayground,
    Competition,
    Data,
    # Preprocessor
    upload_preprocessor,
    import_preprocessor,
    export_preprocessor,
    download_data
]
